#include <windows.h>
#include <vorbis/vorbisfile.h>
#include "cool_vorbis.h"
#include "filters.h"

typedef struct {
    OggVorbis_File vf;
    char sample_buffer[CHUNKSIZE * 2];
    int bpos, bwaiting;
    int nch;
} INPUT;

static size_t read_func ( void *ptr, size_t size, size_t nmemb, HANDLE src )
{
    DWORD br = 0;
    ReadFile ( src, ptr, size*nmemb, &br, 0 );
    return br;
}

static int seek_func ( HANDLE f, ogg_int64_t offset, int whence )
{
    SetFilePointer ( f, (DWORD)offset, (long *)&offset+1, whence );
    return 0;
}

static int close_func (HANDLE f)
{
    CloseHandle ( f );
    return 0;
}

static long tell_func ( HANDLE f )
{
    return SetFilePointer ( f, 0, 0, FILE_CURRENT );
}

static void *callbacks[4] = {
    read_func, seek_func, close_func, tell_func
};

_declspec(dllexport) INPUT* FAR PASCAL OpenFilterInput ( LPSTR fn, long *lSamprate, WORD *wBitsPerSample, WORD *wChannels, HWND hWnd, long *lChunkSize )
{
    INPUT *i;
    vorbis_info *vi;
    HANDLE f = CreateFile ( fn, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0 );
    if ( f == INVALID_HANDLE_VALUE ) return 0;

    i = (INPUT*)LocalAlloc ( LMEM_FIXED | LMEM_ZEROINIT, sizeof(INPUT) );
    if ( !i ) {
        CloseHandle ( f );
        return 0;
    }

    if ( ov_open_callbacks (f, &i->vf, 0, 0, *(ov_callbacks *)&callbacks) ) {
        CloseHandle ( f );
        LocalFree ( i );
        return 0;
    }

    vi = ov_info ( &i->vf, -1 );
    *lSamprate = vi->rate;
    *wBitsPerSample = CV_BPS;
    *wChannels = i->nch = vi->channels;
    *lChunkSize = CHUNKSIZE;

    return i;
}

_declspec(dllexport) void FAR PASCAL CloseFilterInput ( INPUT *i )
{
    ov_clear ( &i->vf );
    LocalFree ( i );
}

_declspec(dllexport) DWORD FAR PASCAL ReadFilterInput ( INPUT *i, BYTE *out, long bytes )
{
    BYTE *buf = out;
    long read = 0L, b;

    do {
        if ( i->bwaiting > 0 ) {
            b = ( i->bwaiting > bytes )  ?  bytes  :  i->bwaiting;
            memcpy ( buf, i->sample_buffer + i->bpos, b );
            buf += b;
            i->bwaiting -= b;
            if ( i->bwaiting == 0 ) i->bpos = 0; else i->bpos += b;
            read += b;
        }

        if ( read < bytes ) {
            int s, c, total_bytes;
            float **samples;
# if CV_BPS == 32
            float *ptr = (float *)i->sample_buffer;
# elif CV_BPS == 16
            short *ptr = (short *)i->sample_buffer;
# elif CV_BPS == 8
            unsigned char *ptr = (unsigned char *)i->sample_buffer;
# else
#error CV_BPS must be 32, 16 or 8
# endif
            int total_samples = ov_read_float ( &i->vf, &samples, CV_SAMPLES, 0 );
            while ( total_samples == OV_HOLE ) {
                total_samples = ov_read_float ( &i->vf, &samples, CV_SAMPLES, 0 );
            }
            if ( total_samples <= 0 ) break;
            total_bytes = total_samples * i->nch * CV_BPS/8;

            for ( s = 0; s < total_samples; s++ ) {
                for ( c = 0; c < i->nch; c++ ) {
# if CV_BPS == 32
                    *ptr++ = samples[c][s] * 32768.f;
# elif CV_BPS == 16
                    float v = samples[c][s] * 32768.f;
                    if      ( v < -32768.f ) *ptr++ = -32768;
                    else if ( v > +32767.f ) *ptr++ = +32767;
                    else                     *ptr++ = (short)v;
# elif CV_BPS == 8
                    float v = samples[c][s] * 128.f;
                    if      ( v < -128.f ) v = -128.f;
                    else if ( v > +127.f ) v = +127.f;
                    *ptr++ = (unsigned char)(v + 128.f);
# else
#error CV_BPS must be 32, 16 or 8
# endif
                }
            }

            if ( read + total_bytes <= bytes ) {
                b = total_bytes;
            } else {
                b = bytes - read;
                i->bpos = b;
                i->bwaiting = total_bytes - i->bpos;
            }

            memcpy ( buf, i->sample_buffer, b );
            buf += b;
            read += b;
        }
    } while ( read < bytes );

    return read;
}

_declspec(dllexport) DWORD FAR PASCAL FilterOptions ( INPUT *i )
{
    return 0;
}       

_declspec(dllexport) long FAR PASCAL FilterGetFileSize ( INPUT *i )
{
    double s = ov_time_total ( &i->vf, -1 );
    if ( s != OV_EINVAL ) {
        vorbis_info *vi = ov_info ( &i->vf, -1 );
        return (long)(s * vi->rate * vi->channels * CV_BPS/8 );
    }
    return 0;
}

_declspec(dllexport) DWORD FAR PASCAL FilterOptionsString ( INPUT *i, char *s )
{
    vorbis_info *vi = ov_info ( &i->vf, -1 );
    vorbis_comment *vc = ov_comment( &i->vf, -1 );
    *s = '\0';
    if ( vi && vc ) {
        wsprintf ( s, "Ogg Vorbis\nAverage bitrate: %u kbps\n%s", (unsigned)((vi->bitrate_nominal+500)/1000), vc->vendor );
    }
    return 0;
}

_declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat ( LPSTR lpszFilename )
{
    if ( lpszFilename && lpszFilename[0] != '\0' ) {
        char header[4];
        DWORD read;
        BOOL error = FALSE;
        HANDLE fh = CreateFile ( lpszFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
        if ( fh == INVALID_HANDLE_VALUE ) return FALSE;
        if ( !ReadFile (fh, header, sizeof(header), &read, NULL) ) error = TRUE;
        CloseHandle ( fh );
        if ( error || read != sizeof (header) ) return FALSE;

        if ( !memcmp (header, "OggS", 4) ) {
            FILE *fp;
            OggVorbis_File vf;
            if ( !(fp = fopen (lpszFilename, "rb")) ) return FALSE;
            if ( ov_open (fp, &vf, NULL, 0) != 0 ) {
                fclose ( fp );
                return FALSE;
            }
            ov_clear ( &vf );
            return TRUE;
        }
    }

    return FALSE;
}
